#!/usr/bin/env python

from brainlab import *
import sys

###############################################################################
#   Create input file template
###############################################################################

#
# Create BRAIN
#
def MakeModel(jobname, c_ee, c_ei, c_ie, p=.1):
    b = BRAIN({'TYPE': 'Brain',
               'JOB': jobname,
               'DURATION':'1.0025',
               'SEED': '-100'})
    del b.parms['OUTPUT_CELLS']
    del b.parms['OUTPUT_CONNECT_MAP']
    del b.parms['DISTANCE']

    lib = b.libs['standard']
    spks = lib['spks']
    comptypes = lib['comptypes']
    syntypes = lib['syntypes']
    sfds = lib['sfds']
    sls = lib['sls']
    spsgs = lib['spsgs']
    celltypes = lib['celltypes']
    
    #
    # Create COLUMN and COLUMN_SHELL
    #
    co_1 = b.COLUMN({'TYPE':'co_1'})
    
    b.AddColumn(co_1)
    
    #
    # Create LAYER and LAYER_SHELL
    #
    ly = b.LAYER({'TYPE':'ly'})
    
    co_1.AddLayerType(ly)
    
    #
    # Create CELL based on 'E' & 'I' from celltypes library
    #
    cl_e = b.Copy(celltypes, 'E', 'cl_e')
    del cl_e.parms['COMPARTMENT']
    
    cl_i = b.Copy(celltypes, 'I', 'cl_i')
    del cl_i.parms['COMPARTMENT']
    
    ly.AddCellType(cl_e, 1000)
    ly.AddCellType(cl_i, 200)
    
    #
    # Create COMPARTMENT based on 'SOMA1' from comptypes library
    #
    cp_e = b.Copy(comptypes, 'SOMA1', 'cp_e')
    del cp_e.parms['CHANNEL']
    del cp_e.parms['SPIKE_HALFWIDTH']
    del cp_e.parms['CA_EXTERNAL']
    for p, v in [('TAU_MEMBRANE',         '0.020        0.0007'),
                 ('R_MEMBRANE',           '40000000     600000'),
                 ('LEAK_CONDUCTANCE',     '0.000000025  0'),
                 ('LEAK_REVERSAL',       '-0.070        0.001'),
                 ('VMREST',              '-0.070        0.001'),
                 ('THRESHOLD',           '-0.052        0.001'),
                 ('CA_INTERNAL',          '0            0'),
                 ('CA_SPIKE_INCREMENT',   '0.0000002    0.000000015'),
                 ('CA_TAU',               '0.080        0.001')]:
        cp_e.parms[p] = v
    
    cp_i = b.Copy(comptypes, 'SOMA1', 'cp_i')
    del cp_i.parms['CHANNEL']
    del cp_i.parms['SPIKE_HALFWIDTH']
    del cp_i.parms['CA_EXTERNAL']
    for p, v in [('TAU_MEMBRANE',         '0.010        0.0003'),
                 ('R_MEMBRANE',           '50000000     700000'),
                 ('LEAK_CONDUCTANCE',     '0.00000002   0'),
                 ('LEAK_REVERSAL',       '-0.065        0.001'),
                 ('VMREST',              '-0.065        0.001'),
                 ('THRESHOLD',           '-0.052        0.001'),
                 ('CA_INTERNAL',          '0            0'),
                 ('CA_SPIKE_INCREMENT',   '0            0'),
                 ('CA_TAU',               '0            0')]:
        cp_i.parms[p] = v
    
    cl_e.AddCompartment(cp_e, 0, 0, 0)
    cl_i.AddCompartment(cp_i, 0, 0, 0)
    
    #
    # Create SPIKESHAPE based on 'AP_Hoffman' from spks library
    #
    ss_e = b.Copy(spks, 'AP_Hoffman', 'ss_e')
    ss_e.parms['VOLTAGES'] = "-0.038 -0.035 -0.030 -0.020 -0.007 0.015 0.030 0.020 0.007 -0.008 -0.016 -0.022 -0.028 -0.033 -0.037 -0.040 -0.043 -0.045 -0.047 -0.049 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059 -0.059"
    
    ss_i = b.Copy(spks, 'AP_Hoffman', 'ss_i')
    ss_i.parms['VOLTAGES'] = "-0.038 -0.035 -0.030 -0.020 -0.007 0.015 0.030 0.020 0.007 -0.008 -0.016 -0.022 -0.028 -0.033 -0.037 -0.040 -0.043 -0.045 -0.047 -0.049 -0.060 -0.060 -0.060 -0.060 -0.060 -0.060 -0.060 -0.060 -0.060 -0.060"
    
    cp_e.parms['SPIKESHAPE'] = ss_e
    cp_i.parms['SPIKESHAPE'] = ss_i
    
    #
    # Create SYNAPSE based on 'E' from syntypes library
    #
    sn_ee = b.Copy(syntypes, 'E', 'sn_ee')
    del sn_ee.parms['PREV_SPIKE_RANGE']
    del sn_ee.parms['RSE_INIT']
    for p, v in [('ABSOLUTE_USE', '0.9          0.1'),
                 ('MAX_CONDUCT',  '0.0000000005 0'),
                 ('SYN_REVERSAL', '0            0.002'),
                 ('DELAY',        '0.003        0.0033')]:
        sn_ee.parms[p] = v
    
    sn_ei = b.Copy(syntypes, 'E', 'sn_ei')
    del sn_ei.parms['PREV_SPIKE_RANGE']
    del sn_ei.parms['RSE_INIT']
    for p, v in [('ABSOLUTE_USE', '0.9          0.1'),
                 ('MAX_CONDUCT',  '0.0000000001 0'),
                 ('SYN_REVERSAL', '0            0.002'),
                 ('DELAY',        '0.003        0.0033')]:
        sn_ei.parms[p] = v
    
    sn_ie = b.Copy(syntypes, 'E', 'sn_ie')
    del sn_ie.parms['PREV_SPIKE_RANGE']
    del sn_ie.parms['RSE_INIT']
    for p, v in [('ABSOLUTE_USE', '0.9          0.1'),
                 ('MAX_CONDUCT',  '0.0000000005 0'),
                 ('SYN_REVERSAL', '-0.070       0.002'),
                 ('DELAY',        '0.0015       0.0018')]:
        sn_ie.parms[p] = v
    
    #
    # Create SYN_FACIL_DEPRESS based on 'F0' from sfds library
    #
    sfd_0 = b.Copy(sfds, 'F0', 'sfd_0')
    
    #sfd_D = b.SYN_FACIL_DEPRESS({'TYPE':'sfd_D'})
    #sfd_D.parms['SFD'] = 'DEPR'
    #sfd_D.parms['DEPR_TAU'] = '0.5          0'
    #sfd_D.parms['FACIL_TAU'] =  '0            0'
    
    sn_ee.parms['SFD_LABEL'] = sfd_0
    sn_ei.parms['SFD_LABEL'] = sfd_0
    sn_ie.parms['SFD_LABEL'] = sfd_0
    
    #
    # Create SYN_LEARNING based on '0Hebb' from sls library
    #
    sl_0 = b.SYN_LEARNING({'TYPE':'sl_0'})
    sl_0.parms['LEARNING'] = 'NONE'
    
    sn_ee.parms['LEARN_LABEL'] = sl_0
    sn_ei.parms['LEARN_LABEL'] = sl_0
    sn_ie.parms['LEARN_LABEL'] = sl_0
    
    #
    # Create SYN_PSG bse on 'PSGexcit' from spsgs library
    #
    psg_e = b.Copy(spsgs, 'PSGexcit', 'psg_e')
    psg_e.parms['PSG_FILE'] = '/home/doursat/NCS/EPSC.txt'
    
    psg_i = b.Copy(spsgs, 'PSGinhib', 'psg_i')
    psg_i.parms['PSG_FILE'] = '/home/doursat/NCS/IPSC.txt'
    
    sn_ee.parms['SYN_PSG'] = psg_e
    sn_ei.parms['SYN_PSG'] = psg_e
    sn_ie.parms['SYN_PSG'] = psg_i
    
    #
    # Create STIMULUS and STIMULUS_INJECT
    #
    a = 0.000000001
    b.AddSimpleStim('st_e_up', (co_1, ly, cl_e), 'c', ampstart = a, ampend = a, dur = (.1, .5))
    
    a = -0.000000001
    b.AddSimpleStim('st_e_down', (co_1, ly, cl_e), 'c', ampstart = a, ampend = a, dur = (.8, .9))
    
    #
    # Create REPORT
    #
    #b.AddSimpleReport('Iapp', co_1, 'sc', dur = (0, 1.0025), prob = .01, format = 'EXP')
    b.AddSimpleReport('Ve', (co_1, ly, cl_e), 'v', dur = (0, 1.0025), prob = .01, format = 'EXP')
    b.AddSimpleReport('Vi', (co_1, ly, cl_i), 'v', dur = (0, 1.0025), prob = .05, format = 'EXP')

    # set up connection prob
    if 'CONNECT' in ly.parms:
        ly.parms['CONNECT'] = []
    ly.AddConnect(cl_e, cl_e, sn_ee, prob = p, speed = 1)
    ly.AddConnect(cl_e, cl_i, sn_ei, prob = p, speed = 1)
    ly.AddConnect(cl_i, cl_e, sn_ie, prob = p, speed = 1)

    # set up connection strengths
    sn_ee.parms['MAX_CONDUCT'] = c_ee
    sn_ei.parms['MAX_CONDUCT'] = c_ei
    sn_ie.parms['MAX_CONDUCT'] = c_ie

    return b

###############################################################################
#   Public API
###############################################################################

#def SetConnectionProbability(p):
#    if 'CONNECT' in ly.parms:
#        ly.parms['CONNECT'] = []
#    ly.AddConnect(cl_e, cl_e, sn_ee, prob = p, speed = 1)
#    ly.AddConnect(cl_e, cl_i, sn_ei, prob = p, speed = 1)
#    ly.AddConnect(cl_i, cl_e, sn_ie, prob = p, speed = 1)

#def SetMaxConduct(c_ee, c_ei, c_ie):
#    sn_ee.parms['MAX_CONDUCT'] = c_ee
#    sn_ei.parms['MAX_CONDUCT'] = c_ei
#    sn_ie.parms['MAX_CONDUCT'] = c_ie

###############################################################################
#   Defaults
###############################################################################

#SetConnectionProbability(0.1)
#SetMaxConduct('0.0000000005 0', '0.0000000001 0', '0.0000000005 0')

###############################################################################
#   Main
###############################################################################

if __name__ == "__main__":
    # this code will only be executed when this script is invoked from the command
    # line (rather than being imported into another script)
    
    print "I am saving the .in file to 'wang99_Ne_Ni_test.in' as a demonstration"
    f = open('wang99_Ne_Ni_test.in', 'w')
    f.write(`b`)
    f.close()

#    print "Now I am simulating the model using NCS, this could take a while . . ."
#    b.Run(verbose = True)
#    print "Done"
